#include "../agc.h"
#include <stdio.h>

#ifndef FALSE
#define FALSE (1 == 0)
#endif

#ifndef TRUE
#define TRUE (0 == 0)
#endif

#define AGC_GET(X) *((AGC_WORD**) (X))
#define AGC_SET(X, Y) (AGC_GET(X) = (AGC_WORD*) (Y))

#define IS_MARKED(X) (*((int*) (X)) & 2)
#define MARKED(X) ((AGC_WORD*) (((int) (X)) | 2))
#define UNMARKED(X) ((AGC_WORD*) (((int) (X)) & (~2)))

#define BOX_INTEGER(X) ((AGC_WORD*) (((X) << 2) | 1))
#define UNBOX_INTEGER(X) (((int) (X)) >> 2)

#define IS_VECTOR(X) ((*((int*) (X)) & 15) == 5)
#define VECTOR_LENGTH(X) (*((int*) (X)) >> 4)
#define VECTOR_REF(X, Y) AGC_GET((X) + ((Y) + 1) * sizeof(AGC_WORD*))
#define VECTOR_SET(X, Y, Z) AGC_SET((X) + ((Y) + 1) * sizeof(AGC_WORD*), (Z))

static struct agc *agc = NULL;
static AGC_WORD *zero = NULL, *rootset = NULL;

int is_marked(AGC_WORD *obj) {
	return IS_MARKED(obj);
}

AGC_WORD *copy_marked(struct agc *agc, AGC_WORD *obj) {
	printf("Not copying marked object %x\n", AGC_GET(obj));
	return UNMARKED(AGC_GET(obj));
}

int is_integer(AGC_WORD *obj) {
	return (*((int*) obj) & 3) == 1;
}

AGC_WORD *copy_integer(struct agc *agc, AGC_WORD *obj) {
	AGC_WORD *new = agc_alloc(agc, 1, &rootset);
	printf("Copying integer %d from %x to %x\n",
		UNBOX_INTEGER(AGC_GET(obj)), obj, new);
	AGC_SET(new, AGC_GET(obj));
	AGC_SET(obj, MARKED(new));
	return new;
}

AGC_WORD *make_vector(int length) {
	AGC_WORD *vector = agc_alloc(agc, length + 1, &rootset);

	if(vector) {
		AGC_SET(vector, (length << 4) | 5);
	}
	while(length--) VECTOR_SET(vector, length, zero);

	return vector;
}

int is_vector(AGC_WORD *obj) {
	return IS_VECTOR(obj);
}

AGC_WORD *copy_vector(struct agc *agc, AGC_WORD *obj) {
	int i, length = VECTOR_LENGTH(obj);
	AGC_WORD *new = agc_alloc(agc, length + 1, &rootset);

	printf("Copying vector of length %d from %x to %x\n",
		length, obj, new);

	AGC_SET(new, AGC_GET(obj));
	AGC_SET(obj, MARKED(new));
	for(i = 0; i < length; i++) {
		VECTOR_SET(new, i, agc_copy_object(agc, VECTOR_REF(obj, i)));
	}

	return new;
}

void finalize_vector(AGC_WORD *obj) {
	printf("Finalizing vector at %x\n", obj);
}

int needs_finalization(AGC_WORD *obj) {
	return !IS_MARKED(obj);
}

void update_finalizer(struct agc_finalizer *finalizer) {
	finalizer->obj = UNMARKED(AGC_GET(finalizer->obj));
}

int main(int argc, char **argv) {
	AGC_WORD *x;
	int i;

	agc = agc_init_with_size(16);
	printf("Initialized agc with %d words of memory, %d used\n",
		agc->size, agc->next);

	agc_register_new_type(agc, is_marked, copy_marked);
	printf("Registered new type 'marked'\n");

	agc_register_new_type(agc, is_integer, copy_integer);
	printf("Registered new type 'integer'\n");

	agc_register_new_type(agc, is_vector, copy_vector);
	printf("Registered new type 'vector'\n");

	zero = agc_alloc(agc, 1, &rootset);
	AGC_SET(zero, BOX_INTEGER(0));

	rootset = make_vector(3);
	printf("Allocated rootset; %d words in use\n", agc->next);

	printf("Collecting with no garbage\n");
	agc_collect(agc, &rootset);
	printf("%d words in use after collection\n", agc->next);

	printf("Garbage vector allocated at %x\n",
		make_vector(2));

	x = make_vector(2);
	agc_add_new_finalizer(agc, x, needs_finalization, finalize_vector,
		update_finalizer);
	printf("Finalizer added to garbage vector at %x\n", x);

	x = make_vector(2);
	VECTOR_SET(rootset, 1, x);
	agc_add_new_finalizer(agc, x, needs_finalization, finalize_vector,
		update_finalizer);
	printf("Finalizer added to rooted vector at %x\n", x);

	printf("%d words in use\n", agc->next);
	printf("Allocating garbage to force collection\n");
	for(i = 0; i < 16; i++) {
		agc_alloc(agc, 1, &rootset);
	}
	printf("%d words in use\n", agc->next);

	agc_fini(agc);
	return 0;
}
